#!/usr/bin/env python3

from optparse import OptionParser
import os
import itertools
import logging
import re
import shlex
import subprocess
import traceback
from _gmutils.launcher import Launcher
from _gmutils.xrandr import XRandr
from _gmutils.config import *


def getNextOption(config, section, options):
    """
    Последовательное получение опций options из секции section
    Возвращает список из 2-х элементов - опция и её значение
    """
    for option in options:
        value = config.getString(section, option, None)
        if value is not None:
            yield ["--" + option, value]


def setCommondMode(xRandr, outputs, mode):
    logging.info("Установка общего режима для всех интерфейсов: " + mode)
    for output in outputs:
        try:
            xRandr.run(['--output', output, '--mode', mode])
        except Exception as e:
            logging.warning(str(e))

    return 0


def disableInterface(xRandr, inteface):
    logging.info("Отключение неиспользуемого интерфейса " + inteface)
    try:
        xRandr.run(['--output', inteface, '--off'])
    except Exception as e:
        logging.warning(str(e))


def getPrimaryInterface(config, interfaces):
    for interface in interfaces:
        if config.getString(interface, 'primary', None) == 'True':
            logging.info("Основной интерфейс: " + interface)
            return interface

    logging.warning("Не удалось определить основной интерфейс")
    return ""     


def runApplication(xRandr, output, app):
    if app and shlex.split(app)[:1]:
        pid = 0
        try:
            '''
            На данный момент при использовании xdotool для перемещения и масштабирования artix-view,
            приложение периодически падает. Предположительно проблема связана с chromium. Выпилить
            костыль после исправления проблем с chromium'ом.
            '''
            if shlex.split(app)[:1][0] == "/linuxcash/cash/view/artix-view":
                geometry = xRandr.getInterfaceGeometry(output)
                arg = '%sx%s+%s+%s' % (geometry['width'], geometry['height'], geometry['x'], geometry['y'])
                Launcher(shlex.split(app)[:1][0]).run(shlex.split(app)[1:] + ['-geometry', arg], autologon=False, pipe=False)
                return
            pid = Launcher(shlex.split(app)[:1][0]).run(shlex.split(app)[1:], autologon=False, pipe=False).pid
        except Exception as e:
            logging.error('Ошибка запуска приложения: %s' % repr(e))
            return 1
        geometry = xRandr.getInterfaceGeometry(output)
        Launcher('xdotool').run(
            args=['search', '--sync', '--pid', str(pid), '--name', 'unexisting_application', 'windowsize',
                  geometry['width'], geometry['height'], 'windowmove', geometry['x'], geometry['y']], pipe=False)

def getNode(id):
    logging.debug(f"Получение ноды для устройства '{id}'")
    process = Launcher('xinput').run(['list-props', id])
    out, err = process.communicate()
    if process.returncode != 0:
        logging.error(err)
        return
    node = ''
    for line in filter(lambda x: x, out.split('\n')):
        if re.search('Device Node', line):
            node = re.sub(r'^Device Node.*"(.*)"', r'\1', line.strip())
            break
    return node.strip()

def getBus(id):
    logging.debug(f"Получение адреса шины для устройства '{id}'")
    node = getNode(id)
    process = Launcher('udevadm').run(['info', '--query=property', f'--name={node}'])
    out, err = process.communicate()
    if process.returncode != 0:
        logging.error(err)
        return
    bus = None
    # ex: DEVPATH=/devices/pci0000:00/0000:00:15.0/usb1/1-5/1-5.3/1-5.3:1.0/0003:222A:0001.001F/input/input76/event5
    # ищем 1-5:1.0 или 1-5.3:1.0, или 1-5.3.1:1.0
    pattern=re.compile(r'^DEVPATH=.*(?P<bus>\d+\-\d+(\.\d+)*\:\d+\.\d+).*$')
    for line in filter(lambda x: x, out.split('\n')):
        match = pattern.search(line)
        if match:
            bus = match.group('bus')
            break
    logging.debug(f"Адрес шины для устройства '{id}' - '{bus}'")
    return bus

def getCalibratableInputDevices():
    """ Последовательный вызов xinput_calibrator и cut с параметрами для получения списка калибрируемых устройств
    Returns:
        dict of <str, str> - список калибрируемых устройств <id, название>
    """
    logging.info('Получение списка идентификаторов калибрируемых устройств')
    process = Launcher('xinput_calibrator').run(['--list'])
    out, err = process.communicate()
    pattern = re.compile(r'Device \"(?P<name>.*)\" id=(?P<id>\d+)')
    logging.info('Результат:\n'+ out)
    return {pattern.search(item).group('id'):pattern.search(item).group('name') 
            for item in filter(lambda x: x, out.split('\n')) if pattern.search(item) is not None}

def mapCalibratableInputDevices(config, outputs):
    """ Маппирование калибрируемых устройств на мониторы в зависимости от настроек
    Args:
        config(Config) - объект конфига
        outputs(list of str) - список мониторов
    """
    logging.info('Маппирование калибрируемых устройств')

    # если конфиг калибровки есть, то не будем маппировать устройства
    if os.path.exists('/usr/share/X11/xorg.conf.d/99-calibration.conf'):
        logging.info('Найден конфигурационный файл калибровки')
        return

    outputsMap = {}  # {<имя_монитора>, <список_калибрируемых_устройств>}

    # Старая логика маппирования по имени
    for output in outputs:
        inputs = config.getStringList(output, 'inputs', [])
        if inputs:
            outputsMap[output] = inputs

    if outputsMap:
        mapInputDevicesByName(outputsMap)
        return

    for output in outputs:
        inputs = config.getStringList(output, 'inputsByPath', [])
        for input in inputs:
            outputsMap[input] = output

    # Eсли секции мониторов не содержат настройки маппинга,
    # тогда нужно смаппировать все калибрируемые устройства на primary или первый попавшийся монитор
    if not outputsMap:
        logging.warn('В конфигурационном файле не настроены калибрируемые устройства')
        for id in getCalibratableInputDevices().keys():
            mapDevice(id, XRandr().getPrimaryOutput())
        return

    mapInputDevicesByBus(outputsMap)

def manageInputDeviceBinding(outputs, cashType):
    """ Подключение/отключение устройств ввода в зависимости от типа и настроек кассы
    Args:
        outputs(list of str) - список мониторов 
        cashType(str) - тип кассы (pos/sst-sco)
    """
    logging.info("Управление подключением устройств ввода")
    sstConfig = Config(True, "sst-sco")
    posConfig = Config(True, "pos")
    for output in outputs:
        if cashType == "pos" and sstConfig.getString(output, "unbindDevices", 'False') == 'True':
            for input in sstConfig.getStringList(output, 'inputsByPath', []):
                unbindDevice(input)
        elif cashType == "sst-sco" and posConfig.getString(output, "unbindDevices", 'False') == 'True':
            for input in posConfig.getStringList(output, 'inputsByPath', []):
                unbindDevice(input)
        inputs = sstConfig.getStringList(output, 'inputsByPath', []) \
            if cashType == 'sst-sco' else posConfig.getStringList(output, 'inputsByPath', [])
        for input in inputs:
            bindDevice(input)

def mapInputDevicesByName(outputsMap):
    """ Маппирование калибрируемых устройств на мониторы по имени в зависимости от настроек
    Args:
        outputsMap(str, str list) - <монитор, список названий устройств>
    """
    logging.info("Маппирование калибрируемых устройств на мониторы по имени")
    for output in outputsMap:
        for input in outputsMap[output]:
            mapDevice(input, output)

def mapInputDevicesByBus(outputsMap):
    """ Маппирование калибрируемых устройств на мониторы по адресу шины в зависимости от настроек
    Args:
        outputsMap(str, str) - <адрес шины, монитор>
    """
    logging.info("Маппирование калибрируемых устройств на мониторы по адресу шины")
    for id in getCalibratableInputDevices().keys():
        bus = getBus(id)
        if bus and bus in outputsMap:
            mapDevice(id, outputsMap[bus])

def mapDevice(device, screen):
    """ Маппирование калибрируемых устройств
    Args:
        device(str) - название или id устройстваы
        screen(str) - монитор
    """
    xinputProcess = Launcher('xinput').run(['--map-to-output', device, screen])
    _, err = xinputProcess.communicate()
    if xinputProcess.returncode != 0:
        logging.error(err)

def unbindDevice(bus):
    """ Отключить устройство от драйвера по адресу шины
    Args:
        bus(str) - адрес шины
    """
    logging.debug(f"Отключение устройства '{bus}'")
    if not checkDeviceStatus(bus):
        logging.debug("Уже отключено")
        return
    try:
        with open("/sys/bus/usb/drivers/usbhid/unbind", "w") as file:
            file.write(bus)
    except Exception as e:
        logging.error(e)

def bindDevice(bus):
    """ Подключить устройство к драйверу по адресу шины
    Args:
        bus(str) - адрес шины
    """
    logging.debug(f"Подключение устройства '{bus}'")
    if checkDeviceStatus(bus):
        logging.debug("Уже подключено")
        return
    try:
        with open("/sys/bus/usb/drivers/usbhid/bind", "w") as file:
            file.write(bus)
    except Exception as e:
        logging.error(e)

def checkDeviceStatus(bus):
    """ Проверить, активно ли устройство
    Args:
        bus(str) - адрес шины
    """
    return os.path.exists(os.path.join('/sys/bus/usb/drivers/usbhid', bus))

def addMode(modes, output, mode):
    if mode in modes[output]:
        return

    logging.info('Добавление нового разрешения %s' % mode)
    match = re.search(r'^([0-9]+)x([0-9]+)$', mode)
    if not match:
        logging.error('Некорректный формат разрешения')
        return

    cvtProcess = subprocess.Popen(['cvt', match.group(1), match.group(2), '60'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    tailProcess = subprocess.Popen(['tail', '-n', '1'], stdin=cvtProcess.stdout, stdout=subprocess.PIPE, universal_newlines=True)
    cvtProcess.stdout.close()
    out, err = tailProcess.communicate()
    if tailProcess.returncode != 0:
        logging.error("Ошибка получения описания разрешения: %s" % err)
        return

    modeRaw = out.strip()

    match = re.search(r'"(.*?)"(.*?)$', modeRaw)
    logging.info('match.groups() = %s' % str(match.groups()))
    if not match:
        logging.error("Ошибка получения имени разрешения")
        return

    modeData = filter(lambda x: x, match.group(2).strip().split(' '))

    # в этом месте мы не анализируем ответ от xrandr, т.к. если по счастливой случайности имя режима уже есть,
    # то он вернёт ошибку, которую сложно парсить
    newmodeProcess = Launcher('xrandr').run(args=['--newmode', mode] + list(modeData))
    out, err = newmodeProcess.communicate()
    if newmodeProcess.returncode != 0:
        logging.error("Ошибка создания нового разрешения: %s" % err)
        return

    addmodeProcess = Launcher('xrandr').run(args=['--addmode', output, mode])
    out, err = addmodeProcess.communicate()
    if addmodeProcess.returncode != 0:
        logging.error("Ошибка добавления нового разрешения: %s" % err)


def gmMain():
    logging.basicConfig(filename='/linuxcash/logs/current/graphicsmode.log', level=logging.DEBUG, format='%(asctime)s - %(levelname)s - %(message)s')
    parser = OptionParser()
    parser.add_option("-q", "--query", action="store_true", dest="query", help="Вывести в консоль информацию о подключенных мониторах")
    (options, args) = parser.parse_args()

    xRandr = XRandr()
    logging.info("XRandr получен")
    if options.query:
        xRandr.query()
        exit(0)

    cashType = 'pos'
    if os.path.exists('/linuxcash/cash/data/tmp/startapp'):
        with open('/linuxcash/cash/data/tmp/startapp', 'r') as startapp:
            cashType = startapp.readline().strip()

    config = Config(True, cashType if isPosSco else None)
    if not config.getBool('main', 'setMode', False):
        logging.warning("Установка режима выключена в настройках")
        return 0

    # Если касса не является трансформером, то настройка runAppWith использоваться не должна
    if not isPosSco:
        for section in config.sections():
            if config.hasOption(section, 'runAppWith'):
                config.removeOption(section, 'runAppWith')

    confMonitors = config.sections()
    confMonitors.remove('main')
    modes = xRandr.getInterfacesModes()
    outputs = list(modes.keys())
    logging.info("Интерфейсы в системе: %s" % outputs)

    for confMonitor in confMonitors:
        if (confMonitor not in outputs) and (confMonitor != 'default_'):
            logging.warning("Монитор %s заданный в настройках не подключен" % confMonitor)

    commonArgs = list(itertools.chain.from_iterable(getNextOption(config, 'main', ['fb', 'dpi', 'fbmm'])))
    commonMode = config.getString('main', 'mode', '800x600')

    optList = ['rate', 'reflect', 'rotate', 'set', 'scale', 'scale-from', 'transform', 'panning', 'gamma', 'brightness']
    if not confMonitors:
        return setCommondMode(xRandr, outputs, commonMode)    
    
    for output in outputs:
        if output in confMonitors or output == 'default' or (config.getMatchedRegexSection(output) is not None):
            continue

        disableInterface(xRandr, output) 

    primaryInterface = getPrimaryInterface(config, outputs)
    outputs.remove(primaryInterface)
    outputs = [primaryInterface] + outputs

    for output in outputs:
        base = commonArgs + ['--output', output]
        args = []
        section = config.getMatchedRegexSection(output)

        if (section in confMonitors) or (section == 'default'):
            if 'default_' in confMonitors:
                output = 'default_'
                section = 'default_'

            mode = config.getString(section, 'mode', commonMode)
            addMode(modes, output, mode)

            args.append(['--mode', mode])
            args.extend(list(getNextOption(config, section, optList)))
        else:
            continue

        primary = config.getString(section, 'primary', None) == 'True'

        if primary:
            args.append(['--primary'])
        elif config.getString(section, 'duplicate', None) == 'True':
            logging.info("Интерфейс %s запущен в режиме дублирования %s" % (output, primaryInterface))
            args.append(['--same-as', primaryInterface])
        else:
            args.append(['--right-of', primaryInterface])

        logging.info("Настройка интерфейса " + output)
        try:
            xRandr.run(base + list(itertools.chain.from_iterable(args)))
        except Exception as e:
            logging.warning(str(e))
            for opt in args:  # если пошло что-то не так и целая команда не выполнилась, то пробуем выполнить по частям
                try:
                    xRandr.run(base + opt)
                except Exception as e:
                    logging.warning(str(e))
        logging.info("Режим успешно установлен")

        if not primary and (not config.hasOption(section, 'runAppWith') or cashType == config.getString(section, 'runAppWith', '')):
            runApplication(xRandr, output, config.getString(section, 'application', None))

    mapCalibratableInputDevices(config, outputs)
    if isPosSco:
        manageInputDeviceBinding(outputs, cashType)

    return 0


if __name__ == "__main__":
    try:
        gmMain()
    except:
        logging.error(traceback.format_exc())
